'use strict';

function decodeDownlink(input) {
    const hexStr = byte2HexString(input.bytes);
    const obj = payloadParser(hexStr);

    return {
        data: obj,
        warnings: [],
        errors: []
    };
}

/**
 * This is the function to create a payload object by decoding hex string
 * @param {String} hexStr
 * @return {Object}
 */
function payloadParser(hexStr) {
    const LoRaMessgeType = {
        0: 'Unconfirmed',
        1: 'Confirmed'
    };
    const SonicBeamWidth = {
        254: 'Small',
        253: 'Medium',
        252: 'Wide',
        55: 'User-defined',
    };
    const SonicBurstLength = {
        5: 'Normal',
        3: 'Short',
        16: 'Very short',
    };
    const SonicTransmittingPower = {
        63: 'High',
        40: 'Medium',
        12: 'Low'
    };
    const SonicSensitivity = {
        63: 'Maximum',
        51: 'Very high',
        48: 'High',
        38: 'Medium',
        27: 'Low',
        15: 'Minimum',
    };
    const SonicEvaluationMethod = {
        1: 'Average value'
    };
    const SonicApplicationFilter = {
        1: 'Container filling'
    };
    const ValveOpenSignal = {
        1: 'Sensor 1',
        2: 'Sensor 2'
    };
    const ValveTriggerEventType = {
        1: 'State change',
        2: 'Valve open',
        3: 'Valve closed',
    };
    const NodeOutputLogic = {
        1: 'Normally open',
        2: 'Normally closed'
    };
    const NodeTriggerEventType = {
        1: 'State change',
        2: 'Target detected',
        3: 'No target detected',
    };
    const GPSAccuracyMode = {
        1: 'Eco mode',
        2: 'Precision mode',
    };
    var len;
    var sID;
    var obj = {};

    obj.payload = hexStr;
    for (var i = 0; i < hexStr.length; i = i + 2) {
        len = parseInt(hexStr.substr(i, 2), 16);
        sID = hexStr.substr(i + 2, 4);

        if (sID == 'F000') {
            obj.all_param_inquired = 'true';
        }

        // Downlink: Device Config
        else if (sID == 'F100') {
            obj.all_device_param_inquired = 'true';
        }
        else if (sID == 'F101') {
            if (len == 2) {
                obj.ble_enable_inquired = 'true';
            }
            else {
                obj.ble_enable = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F102') {
            if (len == 2) {
                obj.device_name_inquired = 'true';
            }
            else {
                obj.device_name = hex2string(hexStr.substr(i + 6, 32));
            }
        }
        else if (sID == 'F103') {
            obj.reset_counter = 'required';
        }
        else if (sID == 'F104') {
            obj.factory_reset = 'required';
        }
        else if (sID == 'F105') {
            obj.change_password = 'required';
        }
        else if (sID == 'F106') {
            obj.change_password_1 = 'required';
        }
        else if (sID == 'F107') {
            obj.change_password_2 = 'required';
        }

        // Downlink: GPS Config
        else if (sID == 'F200') {
            obj.gps_param_inquired = 'true';
        }
        else if (sID == 'F201') {
            if (len == 2) {
                obj.gps_acquisition_active_inquired = 'true';
            }
            else {
                obj.gps_acquisition_active = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F202') {
            if (len == 2) {
                obj.gps_acquisition_interval_inquired = 'true';
            }
            else {
                obj.gps_acquisition_interval = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }
        else if (sID == 'F203') {
            if (len == 2) {
                obj.gps_next_acquisition_inquired = 'true';
            }
            else {
                obj.gps_next_acquisition = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }
        else if (sID == 'F204') {
            if (len == 2) {
                obj.gps_accuracy_mode_inquired = 'true';
            }
            else {
                const gpsAccuracyModeIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.gps_accuracy_mode = (GPSAccuracyMode[gpsAccuracyModeIndex] ? GPSAccuracyMode[gpsAccuracyModeIndex] : 'Invalid');    
            }
        }
        else if (sID == 'F205') {
            if (len == 2) {
                obj.gps_message_type_inquired = 'true';
            }
            else {
                const messageTypeIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.gps_message_type = (LoRaMessgeType[messageTypeIndex] ? LoRaMessgeType[messageTypeIndex] : 'Invalid');
            }
        }
        else if (sID == 'F206') {
            if (len == 2) {
                obj.gps_number_of_transmission_inquired = 'true';
            }
            else {
                obj.gps_number_of_transmission = parseInt(hexStr.substr(i + 6, 2), 16);
            }
        }
        else if (sID == 'F207') {
            obj.gps_localization = 'required';
        }

        // Downlink: LoRa Config
        else if (sID == 'F300') {
            obj.lora_param_inquired = 'true';
        }
        else if (sID == 'F301') {
            if (len == 2) {
                obj.lora_message_type_inquired = 'true';
            }
            else {
                const messageTypeIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.lora_message_type = (LoRaMessgeType[messageTypeIndex] ? LoRaMessgeType[messageTypeIndex] : 'Invalid');
            }
        }
        else if (sID == 'F302') {
            if (len == 2) {
                obj.lora_number_of_transmission_inquired = 'true';
            }
            else {
                obj.lora_number_of_transmission = parseInt(hexStr.substr(i + 6, 2), 16);
            }
        }
        else if (sID == 'F303') {
            if (len == 2) {
                obj.lora_spreading_factor_inquired = 'true';
            }
            else {
                const spreadingFactor = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.lora_spreading_factor = (spreadingFactor === 255 ? 'ADR' : spreadingFactor);
            }
        }
        else if (sID == 'F304') {
            if (len == 2) {
                obj.lora_data_transmission_active_inquired = 'true';
            }
            else {
                obj.lora_data_transmission_active = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F305') {
            if (len == 2) {
                obj.lora_transmission_interval_inquired = 'true';
            }
            else {
                obj.lora_transmission_interval = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }
        else if (sID == 'F306') {
            if (len == 2) {
                obj.lora_next_transmission_inquired = 'true';
            }
            else {
                obj.lora_next_transmission = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }
        else if (sID == 'F307') {
            if (len == 2) {
                obj.lora_downlink_config_active_inquired = 'true';
            }
            else {
                obj.lora_downlink_config_active = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F308') {
            if (len == 2) {
                obj.lora_downlink_config_ack_active_inquired = 'true';
            }
            else {
                obj.lora_downlink_config_ack_active = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F309') {
            if (len == 2) {
                obj.lora_sub_band_inquired = 'true';
            }
            else {
                const subBand = parseInt(hexStr.substr(i + 6, 8), 16);
                obj.lora_sub_band = (subBand === 255) ? 'All' : subBand;
            }
        }

        // Downlink: UltraSonic Config
        else if (sID == 'F400') {
            obj.us_param_inquired = 'true';
        }
        else if (sID == 'F401') {
            if (len == 2) {
                obj.us_beam_width_inquired = 'true';
            }
            else {
                const SonicBeamWidthVal = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.us_beam_width = (SonicBeamWidth[SonicBeamWidthVal] ? SonicBeamWidth[SonicBeamWidthVal] : 'Invalid');
            }
        }
        else if (sID == 'F402') {
            if (len == 2) {
                obj.us_user_defined_burst_length_inquired = 'true';
            }
            else {
                const SonicBurstLengthVal = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.us_user_defined_burst_length = (SonicBurstLength[SonicBurstLengthVal] ? SonicBurstLength[SonicBurstLengthVal] : 'Invalid');
            }
        }
        else if (sID == 'F403') {
            if (len == 2) {
                obj.us_user_defined_transmitting_power_inquired = 'true';
            }
            else {
                const SonicTransmittingPowerVal = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.us_user_defined_transmitting_power = (SonicTransmittingPower[SonicTransmittingPowerVal] ? SonicTransmittingPower[SonicTransmittingPowerVal] : 'Invalid');
            }
        }
        else if (sID == 'F404') {
            if (len == 2) {
                obj.us_user_defined_sensitivity_inquired = 'true';
            }
            else {
                const SonicSensitivityVal = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.us_user_defined_sensitivity = (SonicSensitivity[SonicSensitivityVal] ? SonicSensitivity[SonicSensitivityVal] : 'Invalid');
            }
        }
        else if (sID == 'F405') {
            if (len == 2) {
                obj.us_full_distance_inquired = 'true';
            }
            else {
                obj.us_full_distance = parseInt(hexStr.substr(i + 6, 4), 16);
            }
        }
        else if (sID == 'F406') {
            if (len == 2) {
                obj.us_empty_distance_inquired = 'true';
            }
            else {
                obj.us_empty_distance = parseInt(hexStr.substr(i + 6, 4), 16);
            }
        }
        else if (sID == 'F407') {
            if (len == 2) {
                obj.us_measurement_sequence_active_inquired = 'true';
            }
            else {
                obj.us_measurement_sequence_active = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F408') {
            if (len == 2) {
                obj.us_evaluation_method_inquired = 'true';
            }
            else {
                const evaluationMethodVal = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.us_evaluation_method = (SonicEvaluationMethod[evaluationMethodVal] ? SonicEvaluationMethod[evaluationMethodVal] : 'Invalid');
            }
        }
        else if (sID == 'F409') {
            if (len == 2) {
                obj.us_measurements_per_sequence_inquired = 'true';
            }
            else {
                obj.us_measurements_per_sequence = parseInt(hexStr.substr(i + 6, 2), 16);
            }
        }
        else if (sID == 'F40A') {
            if (len == 2) {
                obj.us_at_intervals_of_inquired = 'true';
            }
            else {
                obj.us_at_intervals_of = parseInt(hexStr.substr(i + 6, 4), 16);
            }
        }
        else if (sID == 'F40B') {
            if (len == 2) {
                obj.us_application_filter_active_inquired = 'true';
            }
            else {
                obj.us_application_filter_active = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F40C') {
            if (len == 2) {
                obj.us_post_processing_filter_inquired = 'true';
            }
            else {
                const filterVal = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.us_post_processing_filter = (SonicApplicationFilter[filterVal] ? SonicApplicationFilter[filterVal] : 'Invalid');
            }
        }
        else if (sID == 'F40D') {
            if (len == 2) {
                obj.us_additional_measurement_inquired = 'true';
            }
            else {
                obj.us_additional_measurement = parseInt(hexStr.substr(i + 6, 2), 16);
            }
        }
        else if (sID == 'F40E') {
            if (len == 2) {
                obj.water_body_level_active_inquired = 'true';
            }
            else {
                obj.water_body_level_active = parseInt(hexStr.substr(i + 6, 2), 16);
            }
        }
        else if (sID == 'F40F') {
            if (len == 2) {
                obj.distance_to_water_body_ground_inquired = 'true';
            }
            else {
                obj.distance_to_water_body_ground = parseInt(hexStr.substr(i + 6, 4), 16);
            }
        }

        // Downlink: Device Information
        else if (sID == 'F500') {
            obj.device_info_inquired = 'true';
        }
        else if (sID == 'F501') {
            if (len == 2) {
                obj.part_number_inquired = 'true';
            }
            else {
                obj.part_number = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }
        else if (sID == 'F502') {
            if (len == 2) {
                obj.hardware_revision_inquired = 'true';
            }
            else {
                const major = parseInt(hexStr.substr(i + 6, 4), 16);
                const minor = parseInt(hexStr.substr(i + 10, 4), 16);
                const patch = parseInt(hexStr.substr(i + 14, 4), 16);
                obj.hardware_revision = major + '.' + minor + '.' + patch;
            }
        }
        else if (sID == 'F503') {
            if (len == 2) {
                obj.firmware_revision_inquired = 'true';
            } else {
                const major = parseInt(hexStr.substr(i + 6, 4), 16);
                const minor = parseInt(hexStr.substr(i + 10, 4), 16);
                const patch = parseInt(hexStr.substr(i + 14, 4), 16);
                obj.firmware_revision = major + '.' + minor + '.' + patch;
            }
        }

        // Downlink ACK: Valve(UCC) Config
        else if (sID == 'F600') {
            obj.valve_param_inquired = 'true';
        }
        else if (sID == 'F601') {
            if (len == 2) {
                obj.valve_monitoring_message_type_inquired = 'true';
            } else {
                const messageTypeIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.valve_monitoring_message_type = (LoRaMessgeType[messageTypeIndex] ? LoRaMessgeType[messageTypeIndex] : 'Invalid');
            }
        }
        else if (sID == 'F602') {
            if (len == 2) {
                obj.valve_monitoring_interval_inquired = 'true';
            } else {
                obj.valve_monitoring_interval = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }

        // Downlink ACK: Valve Config
        else if (sID == 'F700') {
            obj.valve_param_inquired = 'true';
        }
        else if (sID == 'F701') {
            if (len == 2) {
                obj.valve_1_open_signal_inquired = 'true';
            } else {
                const openSignalIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.valve_1_open_signal = (ValveOpenSignal[openSignalIndex] ? ValveOpenSignal[openSignalIndex] : 'Invalid');
            }
        }
        else if (sID == 'F702') {
            if (len == 2) {
                obj.valve_1_event_driven_transmission_inquired = 'true';
            } else {
                obj.valve_1_event_driven_transmission = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F703') {
            if (len == 2) {
                obj.valve_1_trigger_event_inquired = 'true';
            } else {
                const triggerEventIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.valve_1_trigger_event = (ValveTriggerEventType[triggerEventIndex] ? ValveTriggerEventType[triggerEventIndex] : 'Invalid');
            }
        }
        else if (sID == 'F704') {
            if (len == 2) {
                obj.valve_1_message_type_inquired = 'true';
            } else {
                const messageTypeIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.valve_1_message_type = (LoRaMessgeType[messageTypeIndex] ? LoRaMessgeType[messageTypeIndex] : 'Invalid');
            }
        }
        else if (sID == 'F705') {
            if (len == 2) {
                obj.valve_1_num_of_transmission_inquired = 'true';
            } else {
                obj.valve_1_num_of_transmission = parseInt(hexStr.substr(i + 6, 2), 16);
            }
        }
        else if (sID == 'F706') {
            if (len == 2) {
                obj.valve_1_monitoring_interval_inquired = 'true';
            } else {
                obj.valve_1_monitoring_interval = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }
        else if (sID == 'F707') {
            if (len == 2) {
                obj.valve_2_open_signal_inquired = 'true';
            } else {
                const openSignalIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.valve_2_open_signal = (ValveOpenSignal[openSignalIndex] ? ValveOpenSignal[openSignalIndex] : 'Invalid');
            }
        }
        else if (sID == 'F708') {
            if (len == 2) {
                obj.valve_2_event_driven_transmission_inquired = 'true';
            } else {
                obj.valve_2_event_driven_transmission = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F709') {
            if (len == 2) {
                obj.valve_2_trigger_event_inquired = 'true';
            } else {            
                const triggerEventIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.valve_2_trigger_event = (ValveTriggerEventType[triggerEventIndex] ? ValveTriggerEventType[triggerEventIndex] : 'Invalid');
            }
        }
        else if (sID == 'F70A') {
            if (len == 2) {
                obj.valve_2_message_type_inquired = 'true';
            } else {   
                const messageTypeIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.valve_2_message_type = (LoRaMessgeType[messageTypeIndex] ? LoRaMessgeType[messageTypeIndex] : 'Invalid');
            }
        }
        else if (sID == 'F70B') {
            if (len == 2) {
                obj.valve_2_num_of_transmission_inquired = 'true';
            } else {   
                obj.valve_2_num_of_transmission = parseInt(hexStr.substr(i + 6, 2), 16);
            }
        }
        else if (sID == 'F70C') {
            if (len == 2) {
                obj.valve_2_monitoring_interval_inquired = 'true';
            } else {   
                obj.valve_2_monitoring_interval = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }

        // Downlink ACK: Node Config
        else if (sID == 'F800') {
            obj.node_param_inquired = 'true';
        }
      	else if (sID == 'F801') {
            if (len == 2) {
                obj.node_1_output_logic_inquired = 'true';
            } else {
                const outputLogicIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.node_1_output_logic = (NodeOutputLogic[outputLogicIndex] ? NodeOutputLogic[outputLogicIndex] : 'Invalid');
            }
        }
        else if (sID == 'F802') {
            if (len == 2) {
                obj.node_1_event_driven_transmission_inquired = 'true';
            } else {
                obj.node_1_event_driven_transmission = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F803') {
            if (len == 2) {
                obj.node_1_trigger_event_inquired = 'true';
            } else {
                const triggerEventIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.node_1_trigger_event = (NodeTriggerEventType[triggerEventIndex] ? NodeTriggerEventType[triggerEventIndex] : 'Invalid');
            }
        }
        else if (sID == 'F804') {
            if (len == 2) {
                obj.node_1_message_type_inquired = 'true';
            } else {
                const messageTypeIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.node_1_message_type = (LoRaMessgeType[messageTypeIndex] ? LoRaMessgeType[messageTypeIndex] : 'Invalid');
            }
        }
        else if (sID == 'F805') {
            if (len == 2) {
                obj.node_1_num_of_transmission_inquired = 'true';
            } else {
                obj.node_1_num_of_transmission = parseInt(hexStr.substr(i + 6, 2), 16);
            }
        }
        else if (sID == 'F806') {
            if (len == 2) {
                obj.node_1_monitoring_interval_inquired = 'true';
            } else {
                obj.node_1_monitoring_interval = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }
        else if (sID == 'F807') {
            if (len == 2) {
                obj.node_2_output_logic_inquired = 'true';
            } else {
                const outputLogicIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.node_2_output_logic = (NodeOutputLogic[outputLogicIndex] ? NodeOutputLogic[outputLogicIndex] : 'Invalid');
            }
        }
        else if (sID == 'F808') {
            if (len == 2) {
                obj.node_2_event_driven_transmission_inquired = 'true';
            } else {
                obj.node_2_event_driven_transmission = (parseInt(hexStr.substr(i + 6, 2), 16) > 0);
            }
        }
        else if (sID == 'F809') {
            if (len == 2) {
                obj.node_2_trigger_event_inquired = 'true';
            } else {
                const triggerEventIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.node_2_trigger_event = (NodeTriggerEventType[triggerEventIndex] ? NodeTriggerEventType[triggerEventIndex] : 'Invalid');
            }
        }
        else if (sID == 'F80A') {
            if (len == 2) {
                obj.node_2_message_type_inquired = 'true';
            } else {
                const messageTypeIndex = parseInt(hexStr.substr(i + 6, 2), 16);
                obj.node_2_message_type = (LoRaMessgeType[messageTypeIndex] ? LoRaMessgeType[messageTypeIndex] : 'Invalid');
            }
        }
        else if (sID == 'F80B') {
            if (len == 2) {
                obj.node_2_num_of_transmission_inquired = 'true';
            } else {
                obj.node_2_num_of_transmission = parseInt(hexStr.substr(i + 6, 2), 16);
            }
        }
        else if (sID == 'F80C') {
            if (len == 2) {
                obj.node_2_monitoring_interval_inquired = 'true';
            } else {
                obj.node_2_monitoring_interval = parseInt(hexStr.substr(i + 6, 8), 16);
            }
        }

        i = i + (len * 2);
    }
    return obj;
}

function byte2HexString(bytes) {
    var retHexString = '';
    for (var i_b = 0; i_b < bytes.length; i_b++) {
        retHexString = retHexString.concat(('0' + (Number(bytes[i_b]).toString(16))).slice(-2).toUpperCase());
    }
    return retHexString;
}

function number2Int32(value) {
    if (value > 0x7FFFFFFF) {
        return (value - 0x100000000);
    }
    return value;
}

function hex2string(hexx) {
    const hex = hexx.toString();
    var str = '';
    for (var j = 0;
        (j < hex.length && hex.substr(j, 2) !== '00'); j += 2)
        str += String.fromCharCode(parseInt(hex.substr(j, 2), 16));
    return str;
}

function hex2float(hexstring) {
    var bytes = [];
    bytes[0] = parseInt(hexstring.substr(0, 2), 16);
    bytes[1] = parseInt(hexstring.substr(2, 2), 16);
    bytes[2] = parseInt(hexstring.substr(4, 2), 16);
    bytes[3] = parseInt(hexstring.substr(6, 2), 16);
    return decodeFloat(bytes, 1, 8, 23, -126, 127, false);
}

function decodeFloat(bytes, signBits, exponentBits, fractionBits, eMin, eMax, littleEndian) {
    var binary = '';
    for (var z = 0, l = bytes.length; z < l; z += 1) {
        var bits = bytes[z].toString(2);
        while (bits.length < 8) {
            bits = '0' + bits;
        }
        if (littleEndian) {
            binary = bits + binary;
        } else {
            binary += bits;
        }
    }
    var sign = (binary.charAt(0) === '1') ? -1 : 1;
    var exponent = parseInt(binary.substr(signBits, exponentBits), 2) - eMax;
    var significandBase = binary.substr(signBits + exponentBits, fractionBits);
    var significandBin = '1' + significandBase;
    var cnt = 0;
    var val = 1;
    var significand = 0;
    if (exponent == -eMax) {
        if (significandBase.indexOf('1') === -1)
            return 0;
        else {
            exponent = eMin;
            significandBin = '0' + significandBase;
        }
    }
    while (cnt < significandBin.length) {
        significand += val * parseInt(significandBin.charAt(cnt));
        val = val / 2;
        cnt += 1;
    }
    return sign * significand * Math.pow(2, exponent);
}
